using UnityEngine;
using System.Collections;

/**
 * This script is responsible for the overall flow of the game
 *  - spawning things
 *  - checking game state, winning, losing 
 *  - etc.
 */
public class Gameflow : MonoBehaviour {
	
	public int supremaCount;
	public int supremaCircleRadius;
	public int supremaCircleHeight;
	public GameObject supremumPrefab;
	
	private GameObject[] suprema;
	
	void Start () {
		suprema = new GameObject[supremaCount*2];
		spawnSupremaCircle();
		PlayerPrefs.SetInt("points", 0);
	}
	
	void Update () {
		// Reload suprema on "R"
		if(Input.GetButtonUp("R")){
			Messenger<int>.Broadcast("add time", -10);
			Messenger<string>.Broadcast("play sound", "reload");
			newSuprema();
		}
	}
	
	/*
	 * Destroy all existing suprema and spawn new ones
	 */
	private void newSuprema(){
		
		//foreach(GameObject g in suprema){
		//	Destroy(g);
		//}
		Supremum[] sups = (Supremum[]) FindObjectsOfType(typeof(Supremum));
		foreach(Supremum s in sups){
			Destroy(s.gameObject);
		}
		suprema = new GameObject[supremaCount*2];
		spawnSupremaCircle();
	}
	
	private void spawnSupremaCircle(){
		for(int n = 0 ; n < supremaCount ; n++){
			float val = 360 * n / supremaCount;
			float x = Mathf.Cos(val * Mathf.Deg2Rad);
			float z = Mathf.Sin(val * Mathf.Deg2Rad);
			
			float posx = x * supremaCircleRadius + transform.position.x;
			float posz = z * supremaCircleRadius + transform.position.z;
			
			float randomHeight = Random.Range(-7f, 7f);
						
			Vector3 position = new Vector3(posx, supremaCircleHeight + randomHeight, posz);
			Vector3 direction = (transform.position - position);
			Quaternion lookRotation = Quaternion.LookRotation(direction);
			Quaternion randomRotation = Quaternion.Euler(new Vector3(0,0,Random.Range(0,360)));
			lookRotation *= randomRotation;
			suprema[n] = (GameObject) Instantiate(supremumPrefab, position, lookRotation);
		}
			
		for(int n = 0 ; n < supremaCount ; n++){
			float val = 360 * n / (supremaCount) + 20;
			float x = Mathf.Cos(val * Mathf.Deg2Rad);
			float z = Mathf.Sin(val * Mathf.Deg2Rad);
			
			float posx = x * supremaCircleRadius + transform.position.x;
			float posz = z * supremaCircleRadius + transform.position.z;
			
			Vector3 position = new Vector3(posx, supremaCircleHeight + 25, posz);
			Vector3 direction = (transform.position - position);
			Quaternion lookRotation = Quaternion.LookRotation(direction);
			
			suprema[n+supremaCount] = (GameObject) Instantiate(supremumPrefab, position, lookRotation);
		}
		//*/
		
		// We've placed all the suprema, tell them to spawn their shapes
		Messenger.Broadcast("spawn shapes");
	}
}
