using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class SupSelector : MonoBehaviour {
	
	public Transform[] plates; 
	private Transform[] selectedItems = new Transform[3];

	void OnEnable() {
		Messenger<Transform>.AddListener("select", selecting);
	}
	
	void OnDestroy() {
        Messenger<Transform>.RemoveListener("select", selecting);
    }
	
	// Use this for initialization
	void Start () {
		
	}
	
	// Update is called once per frame
	void Update () {
		if(Input.GetMouseButtonUp(1)){
			Messenger<string>.Broadcast("play sound", "return");
			clearSelection();
		}
	}
	
	private void clearSelection(){
		// Clear selected
			foreach(Transform item in selectedItems){
				if(item != null){
					// Enable the collider and the renderers again
					BoxCollider selectComponent = item.GetComponentInChildren<BoxCollider>();
					if(selectComponent != null)
						selectComponent.enabled = true;
					
					MeshRenderer[] renderers = item.GetComponentsInChildren<MeshRenderer>();
					foreach(MeshRenderer m in renderers){
						m.enabled = true;
					}
				}
			}
			selectedItems = new Transform[3];
			foreach(Transform t in plates){
				if(t.GetChildCount() > 1)
					Destroy(t.GetChild(1).gameObject);
			}
	}
	
	/**
	 * Called when the player 'selects' a supremum, we move a copy
	 * of the selected supremum to to any free page in the player's
	 * hands.
	 */
	public void selecting(Transform selected){
		for(int n = 0 ; n < selectedItems.Length ; n++){
			if(selectedItems[n] == null){
				selectedItems[n] = selected;
				
				// Disable the 'selectable' so we can't choose the same item more than once
				BoxCollider selectComponent = selected.GetComponentInChildren<BoxCollider>();
				if(selectComponent != null)
					selectComponent.enabled = false;
								
				Transform miniCopy = (Transform) Object.Instantiate(selected);
				miniCopy.parent = plates[n];
				miniCopy.transform.position = plates[n].transform.position;
				miniCopy.transform.rotation = plates[n].transform.rotation * Quaternion.Euler(new Vector3(0, 90, 0));
				miniCopy.transform.localScale -= new Vector3(0.99f, 0.99f, 0.99f);
								
				// Disable the renderers on the shapes
				MeshRenderer[] renderers = selected.GetComponentsInChildren<MeshRenderer>();
				foreach(MeshRenderer m in renderers){
					m.enabled = false;
				}
				// Play sound
				Messenger<string>.Broadcast("play sound", "select");
				break;
			}
		}
		
		// Handle a match if we have one
		if(haveThree()){
			checkForMatch();
		}
	}
	
	private void checkForMatch(){
		// The values of the properties can be 1, 2, or 5
		// so three matching items will add up to one of the following
		List<int> matchValues = new List<int>{3, 6, 8, 15};
		
		Supremum[] suprema = new Supremum[3];
		
		int combinedColours = 0;
		int combinedShapes = 0;
		int combinedCount = 0;
		
		for(int n = 0 ; n < suprema.Length ; n++){
			suprema[n] = (Supremum) selectedItems[n].GetComponent("Supremum");
			suprema[n].dumpInfo();
			combinedColours += suprema[n].getColourVal();
			combinedShapes += suprema[n].getShapeVal();
			combinedCount += suprema[n].getCountVal();
		}

		if(matchValues.Contains(combinedColours) && matchValues.Contains(combinedShapes) && matchValues.Contains(combinedCount)){
			match();
		} else {
			noMatch();
		}
	}
	
	/*
	 * With no match just notify the player
	 */
	private void noMatch(){
		Debug.Log("Not a match...");
	}
	
	/*
	 * When there's a match, add points, remove the matching set, 
	 * clear the pages and continue.
	 */
	private void match(){
		Messenger<int>.Broadcast("point", 1);
		Messenger<int>.Broadcast("add time", 5);
		foreach(Transform t in selectedItems){
			Destroy(t.gameObject);
		}
		clearSelection();
	}
	
	private bool haveThree(){
		for(int n = 0 ; n < selectedItems.Length ; n++){
			if(selectedItems[n] == null){
				return false;
			}
		}
		return true;
	}
}
