using UnityEngine;
using System.Collections;

/**
 * A supremum represents a single selectable object. It contains 1-3 similar items 
 * from a selection of three different shapes and three different colours.
 * 
 * On creation its features are randomized.
 * 
 * It also contains an initially invisible "frame" that shows as a sort of HUD when
 * the player looks directly at the supremum.
 */
public class Supremum : MonoBehaviour {
	
	public Transform frame;
	public Transform[] hinges;
	
	public Material[] colors;
	public GameObject[] shapes;
	
	private int count;
	private int color;
	private int shape;
	
	private GameObject[] instShapes = new GameObject[3];
	
	void OnEnable() {
		Messenger.AddListener("spawn shapes", spawnShapes);
	}
	
	void OnDestroy() {
        Messenger.RemoveListener("spawn shapes", spawnShapes);
    }
	
	// Use this for initialization
	void Start () {
		
	}
	
	// Update is called once per frame
	void Update () {
		
	}
	
	public int getShapeVal(){
		int val = shape == 2 ? 5 : shape + 1;
		return val;
	}
	
	public int getColourVal(){
		int val = color == 2 ? 5 : color + 1;
		return val;
	}
	
	public int getCountVal(){
		int val = count-1 == 2 ? 5 : count;
		return val;
	}
	
	public void dumpInfo(){
		Debug.Log("Count: " + count + "  Color: " + color + "  Shape: " + shape);
	}
	
	public void spawnShapes(){
		// How many shapes?
		count = Random.Range(0,3) + 1;
		
		// Which shape?
		shape = Random.Range(0,3);
		
		// Which color?
		color = Random.Range(0,3);
		
		// Instantiate supremum
		for(int n = 0 ; n < count ; n++){
			instShapes[n] = (GameObject) Instantiate(shapes[shape], hinges[n].position, hinges[n].rotation);
			instShapes[n].renderer.material = colors[color];
			instShapes[n].transform.parent = hinges[n];
			
			// Randomize size and rotation a little 
			float randomX = 0f;
			float randomY = 0f;
			float randomRotZ = 0f;
			switch(shape){
			case 0: //circle
				randomX = Random.Range(-2f, 2f);
				randomY = Random.Range(-2f, 2f);
				break;
			case 1: //line
				randomX = Random.Range(-0.05f, 0.5f);
				randomY = Random.Range(-0.1f, 0.1f);
				// Only apply rotation to 1/3
				if(Random.Range(0,3) == 0){
					randomRotZ = Random.Range(-45, 45);
				}
				break;
			case 2: //rect
				randomX = Random.Range(-0.1f, 0.1f);
				randomY = Random.Range(-0.1f, 0.1f);
				// Only apply rotation to 1/3
				if(Random.Range(0,3) == 0){
					randomRotZ = Random.Range(0, 360);
				}
				break;
			}
			instShapes[n].transform.rotation *= Quaternion.Euler(new Vector3(0, 0, randomRotZ));
			instShapes[n].transform.localScale += new Vector3(randomX, randomY, 0);
		}
		
		Fader[] faders = transform.GetComponentsInChildren<Fader>();
		foreach(Fader f in faders){
			f.setOriginalColor();
		}
	}
	
}
